/*
DEBUG BALL HANDLER
Every 1 sec:
- Displays appropriate coloured debug ball depending on unit's suppression state
- Hides balls for unsuppressed or injured units
- Removes balls from dead units (!)
- Places markers on map indicating unit status
*/

tpwcas_fnc_debug = 
	{
	private ["_ball", "_marker", "_level", "_x", "_nul", "_color"];
        
    while { true } do
        {
            {
			if( isNil { _x getVariable "tpwcas_debug_ball" } ) then
				{
				_ball = "Sign_sphere25cm_EP1" createvehicle getposatl _x;  
				_ball setObjectTexture [0,"#(argb,8,8,3)color(0.99,0.99,0.99,0.7,ca)"];  // white
				_ball attachTo [_x,[0,0,2]];   
				_ball hideobject true;  
				
				if( (side _x) getFriend WEST < 0.6 ) then 
					{ _color = "ColorRed"; 
					} 
				else 
					{
					_color = "ColorBlue"; 
					};
				
				_marker = createMarker[ format["tpwcas_markers_%1", _x], position _x];
				_marker setMarkerShape "ICON";
				_marker setMarkerType "mil_triangle";
				_marker setMarkerColor _color;

				_x setVariable ["tpwcas_debug_ball", _ball ];
				_x setVariable ["tpwcas_debug_marker", _marker ];
				}
            else
                {
				_ball = _x getVariable "tpwcas_debug_ball";
				_marker = _x getVariable "tpwcas_debug_marker";
                };

            _marker setMarkerPos (position _x);
                
			if( fleeing _x) then 
				{
				_marker setMarkerType "mil_dot";
				}
            else
                {
				_marker setMarkerType "mil_triangle";
				_marker setMarkerDir (getDir _x);
                };
                
			if( !( isNull _x ) && alive _x) then // better to double check for unit being alive ...
				{
				_level = _x  getvariable ["tpwcas_supstate", 0];
			
				switch ( true ) do
					{
					case ( fleeing _x ): 
						{  
						_ball hideobject false;  
						_ball setObjectTexture [0,"#(argb,8,8,3)color(0.0,0.0,0.0,0.9,ca)"];  // black
						};
					
					case ( _level == 0 ): 
						{  
						_ball hideobject true;  
						};
					
					case ( _level == 1 ): 
						{  
						_ball hideobject false; 
						_ball setObjectTexture [0,"#(argb,8,8,3)color(0.1,0.9,0.1,0.7,ca)"];  // green
						};
					
					case ( _level == 2): 
						{  
						_ball hideobject false; 
						_ball setObjectTexture [0,"#(argb,8,8,3)color(0.9,0.9,0.1,0.7,ca)"]; //yellow
						};
					
					case ( _level == 3 ): 
						{  
						_ball hideobject false; 
						_ball setObjectTexture [0,"#(argb,8,8,3)color(0.9,0.1,0.1,0.7,ca)"]; //red  
						};
					};
				};
			if (lifestate _x != "ALIVE") then
				{  
				_ball hideobject true;  
                };
                
            } foreach allUnits;
            
            {
                if( ! ( isNil { _x getVariable "tpwcas_debug_ball" } ) ) then
                {
                    deleteVehicle ( _x getVariable "tpwcas_debug_ball" );
                    deleteMarker( _x getVariable "tpwcas_debug_marker" );
                };
                
            } foreach allDead;
            sleep 1;
        };
    };